"""Implements the setuptools command 'antlr'."""
import collections
import datetime
import distutils.errors
import distutils.log
import distutils.version
import os.path
import pathlib
import re
import shutil
import shlex
import subprocess
import typing

import setuptools

from setuptools_antlr import __path__
from setuptools_antlr.util import camel_to_snake_case, find_java


class AntlrGrammar(object):
    """Basic information about an ANTLR grammar file.

    For generation of ANTLR based parsers basic information about the grammar like imports is
    necessary. This information and the functionality to retrieve this information out of a grammar
    file is placed in this class.
    """

    def __init__(self, path: pathlib.Path):
        """Initializes a new AntlrGrammar object.

        :param path: path to grammar file
        """
        # by convention grammar name is always equal to file name
        self.name = path.stem
        self.path = path
        self.dependencies = []

    def read_imports(self) -> typing.List[str]:
        """Reads all imported grammars out of grammar file.

        :return: a list of imported grammars
        """
        import_stmt_regex = re.compile('import(.*);')

        try:
            with self.path.open() as f:
                match = import_stmt_regex.search(f.read())
                if match:
                    imported_grammars = match.group(1)
                    return [s.strip() for s in imported_grammars.split(',')]
                else:
                    return []
        except IOError as e:
            raise distutils.errors.DistutilsFileError('Can\'t read grammar "{}"'.format(e.filename))

    def walk(self):
        """Returns dependent grammars by walking the dependency tree of the grammar top-down."""
        for d in self.dependencies:
            yield d
            yield from d.walk()


class ImportGrammarError(Exception):
    """Raised when an imported grammar can't be found in package source directory."""

    def __init__(self, name: str, parent: AntlrGrammar=None):
        """Initializes a new ImportGrammarError object.

        :param name: name of included grammar
        :param parent: parent grammar which includes missing grammar
        """
        super().__init__()
        self.name = name
        self.parent = parent

    def __str__(self):
        """Returns a nicely printable string representation of this ImportGrammarError object.

        :return: a string representation of this error
        """
        return self.name


class AntlrCommand(setuptools.Command):
    """A setuptools command for generating ANTLR based parsers.

    An extra command for setuptools to generate ANTLR based parsers, lexers, listeners and visitors.
    The antlr command wraps the Java based generator provided by ANTLR developers. It
    searches for all grammar files and generates a Python package containing a modules specified in
    the user options. Please keep in mind that only grammars are generated which aren't included by
    other grammars. This prevents generation of shared content like common terminals.

    :cvar _MIN_JAVA_VERSION: Minimal version of java required by ANTLR
    :cvar _EXT_LIB_DIR: Relative path to external libs directory
    :cvar _GRAMMAR_FILE_EXT: File extension of ANTLR grammars
    :cvar description: Description of antlr command
    :cvar user_options: Options which can be passed by the user
    :cvar boolean_options: Subset of user options which are binary
    :cvar negative_opt: Dictionary of user options which exclude each other
    """

    _MIN_JAVA_VERSION = '1.7.0'

    _EXT_LIB_DIR = 'lib'

    _GRAMMAR_FILE_EXT = 'g4'

    description = 'generate a parser based on ANTLR'

    user_options = [
        ('grammars=', 'g', 'specify grammars to generate parsers for'),
        ('output=', 'o', 'specify output directory where all output is generated'),
        ('atn', None, 'generate rule augmented transition network diagrams'),
        ('encoding=', None, 'specify grammar file encoding e.g. euc-jp'),
        ('message-format=', None, 'specify output style for messages in antlr, gnu, vs2005'),
        ('long-messages', None, 'show exception details when available for errors and warnings'),
        ('listener', None, 'generate parse tree listener (default)'),
        ('no-listener', None, 'don\'t generate parse tree listener'),
        ('visitor', None, 'generate parse tree visitor'),
        ('no-visitor', None, 'don\'t generate parse tree visitor (default)'),
        ('depend', None, 'generate file dependencies'),
        ('grammar-options=', None, "set/override a grammar-level options"),
        ('w-error', None, 'treat warnings as error'),
        ('x-dbg-st', None, 'launch StringTemplate visualizer on generated code'),
        ('x-dbg-st-wait', None, 'wait for STViz to close before continuing'),
        ('x-force-atn', None, 'use the ATN simulator for all predictions'),
        ('x-log', None, 'dump lots of logging info to antlr-<timestamp>.log')
    ]

    boolean_options = ['atn', 'long-messages', 'listener', 'no-listener', 'visitor', 'no-visitor', 'depend', 'w-error',
                       'x-dbg-st', 'x-dbg-st-wait', 'x-force-atn', 'x-log']

    negative_opt = {'no-listener': 'listener', 'no-visitor': 'visitor'}

    def initialize_options(self):
        """Sets default values for all the options that this command supports. Note that these
        defaults may be overridden by other commands, by the setup script, by config files, or by
        the command-line.
        """
        self.grammars = None
        self.output = None
        self.atn = 0
        self.encoding = None
        self.message_format = None
        self.long_messages = 0
        self.listener = 1
        self.visitor = 0
        self.depend = 0
        self.grammar_options = {}
        self.w_error = 0
        self.x_dbg_st = 0
        self.x_dbg_st_wait = 0
        self.x_force_atn = 0
        self.x_log = 0

    def finalize_options(self):
        """Sets final values for all the options that this command supports. This is always called
        as late as possible, ie. after any option assignments from the command-line or from other
        commands have been done.
        """
        # find out the output directory if not specified
        self.set_undefined_options('build', ('build_lib', 'output'))

        # parse grammars
        if self.grammars:
            self.grammars = shlex.split(self.grammars, comments=True)

        # parse grammar-level options
        if self.grammar_options:
            tokens = shlex.split(self.grammar_options, comments=True)
            self.grammar_options = dict(t.split('=', 1) for t in tokens)

        # sanity check in case target language is explicitly passed by user
        if 'language' in self.grammar_options:
            if self.grammar_options['language'] != 'Python3':
                raise distutils.errors.DistutilsOptionError('{} isn\'t a supported language. Only Python3 code can be '
                                                            'generated.'.format(self.grammar_options['language']))
        else:
            self.grammar_options['language'] = 'Python3'

        # sanity check for debugging options
        if not self.x_dbg_st and self.x_dbg_st_wait:
            distutils.log.warn('Waiting for StringTemplate visualizer (x_dbg_st_wait) without launching it on '
                               'generated code is enabled (x_dbg_st). Launching of StringTemplate visualizer will be '
                               'forced.')
            self.x_dbg_st = 1

    def _find_antlr(self) -> pathlib.Path:
        """Searches for ANTLR library at setuptools-antlr install location.

        :return: a path to latest ANTLR library or None if library wasn't found
        """
        AntlrJar = collections.namedtuple('AntlrJar', ['file', 'version'])
        antlr_jar_path = pathlib.Path(__path__[0], self._EXT_LIB_DIR)
        antlr_jar_regex = re.compile('^antlr-(\d+(?:.\d+){1,2})-complete.jar$')

        # search for all _files_ matching regex in antlr_jar_path
        antlr_jars = []
        for antlr_jar in antlr_jar_path.iterdir():
            match = antlr_jar_regex.search(antlr_jar.name)
            if antlr_jar_path.joinpath(antlr_jar).is_file() and match:
                version = distutils.version.StrictVersion(match.group(1))
                antlr_jars.append(AntlrJar(antlr_jar, version))

        if antlr_jars:
            # if more than one antlr jar was found return path of the latest version
            latest_antlr_jar = max(antlr_jars, key=lambda x: x.version)
            return pathlib.Path(antlr_jar_path, latest_antlr_jar.file)
        else:
            return None

    @classmethod
    def _find_antlr_log(cls, log_path: pathlib.Path) -> pathlib.Path:
        """Searches for ANTLR log files at passed location.

        :return: a path to the latest ANTLR log file or None if no log file was found
        """
        AntlrLog = collections.namedtuple('AntlrLog', ['file', 'timestamp'])
        antlr_log_regex = re.compile('^antlr-(\d{4}-\d{2}-\d{2}-\d{2}.\d{2}.\d{2}).log$')

        # search for all _files_ matching regex in antlr_log_regex
        antlr_logs = []
        for log_file in log_path.iterdir():
            match = antlr_log_regex.search(log_file.name)
            if log_file.is_file() and match:
                timestamp = datetime.datetime.strptime(match.group(1), '%Y-%m-%d-%H.%M.%S')
                antlr_logs.append(AntlrLog(log_file, timestamp))

        if antlr_logs:
            # if more than one antlr log was found return path of the latest log file
            latest_antlr_log = max(antlr_logs, key=lambda x: x.timestamp)
            return latest_antlr_log.file
        else:
            return None

    def _find_grammars(self, base_path: pathlib.Path=pathlib.Path('.')) -> typing.List[AntlrGrammar]:
        """Searches for all ANTLR grammars in package source directory and returns a list of it.
        Only grammars which aren't included by other grammars are part of this list.

        :param base_path: base path to search for ANTLR grammars
        :return: a list of all found ANTLR grammars
        """
        grammars = []

        def get_grammar(name: str) -> AntlrGrammar:
            """Searches in grammars list for a grammar which has passed name.

            :param name: name of grammar
            :return: an ANTLR grammar
            """
            try:
                return next(g for g in grammars if g.name == name)
            except StopIteration:
                raise ImportGrammarError(name)

        # search for all grammars in package source directory
        for root, _, files in os.walk(str(base_path), followlinks=True):
            grammar_files = [f for f in files if f.endswith("." + self._GRAMMAR_FILE_EXT)]
            for fb in grammar_files:
                grammars.append(AntlrGrammar(pathlib.Path(root, fb)))

        # generate a dependency tree for each grammar
        grammar_tree = []
        try:
            for grammar in grammars:
                imports = grammar.read_imports()
                if imports:
                    try:
                        grammar.dependencies = [get_grammar(i) for i in imports]
                    except ImportGrammarError as e:
                        e.parent = grammar
                        raise
        except ImportGrammarError as e:
            raise distutils.errors.DistutilsFileError('Imported grammar "{}" in file "{}" isn\'t present in package '
                                                      'source directory.'.format(str(e), str(e.parent.path)))
        else:
            # remove all grammars which aren't the root of a dependency tree
            grammar_tree[:] = filter(lambda r: all(r not in g.dependencies for g in grammars),
                                     grammars)

        return grammar_tree

    def run(self):
        """Performs all tasks necessary to generate ANTLR based parsers for all found grammars. This
        process is controlled by the user options passed on the command line or set internally to
        default values.
        """
        java_exe = find_java(self._MIN_JAVA_VERSION)
        if not java_exe:
            raise distutils.errors.DistutilsExecError('no compatible JRE was found on the system')

        antlr_jar = self._find_antlr()
        if not antlr_jar:
            raise distutils.errors.DistutilsExecError('no ANTLR jar was found in lib directory')

        # find grammars and filter result if grammars are passed by user
        grammars = self._find_grammars()
        if self.grammars:
            grammars = filter(lambda g: g.name in self.grammars, grammars)

        # generate parser for each grammar
        for grammar in grammars:
            # build up ANTLR command line
            run_args = [str(java_exe), '-jar', str(antlr_jar)]
            if self.atn:
                run_args.append('-atn')
            if self.encoding:
                run_args.extend(['-encoding', self.encoding])
            if self.message_format:
                run_args.extend(['-message-format', self.message_format])
            if self.long_messages:
                run_args.append('-long-messages')
            run_args.append('-listener' if self.listener else '-no-listener')
            run_args.append('-visitor' if self.visitor else '-no-visitor')
            if self.depend:
                run_args.append('-depend')
            run_args.extend(['-D{}={}'.format(option, value) for option, value in self.grammar_options.items()])
            if self.w_error:
                run_args.append('-Werror')
            if self.x_dbg_st:
                run_args.append('-XdbgST')
            if self.x_dbg_st_wait:
                run_args.append('-XdbgSTWait')
            if self.x_force_atn:
                run_args.append('-Xforce-atn')
            if self.x_log:
                run_args.append('-Xlog')

            # determine location of dependencies e.g. imported grammars and token files
            dependency_dirs = set(g.path.parent for g in grammar.walk())
            if len(dependency_dirs) == 1:
                run_args.extend(['-lib', str(dependency_dirs.pop().absolute())])
            elif len(dependency_dirs) > 1:
                raise distutils.errors.DistutilsOptionError('Imported grammars of \'{}\' are located in more than one '
                                                            'directory. This isn\'t supported by ANTLR. Move all '
                                                            'imported grammars into one'
                                                            'directory.'.format(grammar.name))

            # create package directory
            grammar_dir = grammar.path.parent
            package_dir = pathlib.Path(self.output, grammar_dir, camel_to_snake_case(grammar.name))
            package_dir.mkdir(parents=True, exist_ok=True)
            run_args.extend(['-o', str(package_dir.absolute())])

            grammar_file = grammar.path.name
            run_args.append(str(grammar_file))

            if self.depend:
                dependency_file = pathlib.Path(package_dir, 'dependencies.txt')
                distutils.log.info('generating {} file dependencies -> {}'.format(grammar_file, dependency_file))

                # call ANTLR for file dependency generation
                result = subprocess.run(run_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                        universal_newlines=True, cwd=str(grammar_dir))
                with dependency_file.open('wt') as f:
                    f.write(result.stdout)
            else:
                distutils.log.info('generating {} parser -> {}'.format(grammar.name, package_dir))

                # create Python package
                init_file = pathlib.Path(package_dir, '__init__.py')
                init_file.open('wt').close()

                # call ANTLR for parser generation
                result = subprocess.run(run_args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                                        universal_newlines=True, cwd=str(grammar_dir))
                if result.returncode:
                    raise distutils.errors.DistutilsExecError('{} parser couldn\'t be generated\n'
                                                              '{}'.format(grammar.name, result.stdout))

            # move logging info into build directory
            if self.x_log:
                antlr_log_file = self._find_antlr_log(grammar_dir)
                if antlr_log_file:
                    package_log_file = pathlib.Path(package_dir, antlr_log_file.name)
                    distutils.log.info('dumping logging info of {} -> {}'.format(grammar_file, package_log_file))
                    shutil.move(str(antlr_log_file), str(package_log_file))
                else:
                    distutils.log.warn('no logging info dumped out by ANTLR')



"""Utilities required by 'antlr' setuptools command ."""
import os.path
import pathlib
import shutil
import subprocess
import distutils.version
import re


def camel_to_snake_case(s):
    """Converts a camel cased to a snake cased string.

    :param s: a camel cased string
    :return: a snake cased string
    """
    snake_cased = re.sub('([a-z0-9])([A-Z])', r'\1_\2', re.sub('(.)([A-Z][a-z]+)', r'\1_\2', s)).lower()
    return snake_cased.replace('__', '_')


def validate_java(executable: str, min_java_version: str) -> bool:
    """Validates a Java Runtime Environment (JRE) if it fulfills minimum acceptable version.

    :param executable: Java executable of JRE
    :param min_java_version: minimum acceptable version of Java
    :return: flag whether JRE is at minimum required version
    """
    result = subprocess.run([executable, '-version'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                            universal_newlines=True)

    if result.returncode == 0:
        version_regex = re.compile('\d+(?:.\d+){2}(?:_\d+)')
        version_match = version_regex.search(result.stdout)

        if version_match:
            # create normalized versions containing only valid chars
            validated_version = distutils.version.LooseVersion(version_match.group(0).replace('_', '.'))
            min_version = distutils.version.LooseVersion(min_java_version.replace('_', '.'))

            return validated_version >= min_version

    return False


def find_java(min_java_version: str) -> pathlib.Path:
    """Searches for a working Java Runtime Environment (JRE) set in JAVA_HOME or PATH
    environment variables. A JRE located in JAVA_HOME will be preferred.

    :param min_java_version: minimum acceptable version of Java
    :return: a path to a working JRE or None if no JRE was found
    """
    # first check if a working Java is set in JAVA_HOME
    if 'JAVA_HOME' in os.environ:
        java_bin_dir = os.path.join(os.environ['JAVA_HOME'], 'bin')
        java_exe = shutil.which('java', path=java_bin_dir)
        if java_exe and validate_java(java_exe, min_java_version):
            return pathlib.Path(java_exe)

    # if Java wasn't found in JAVA_HOME fallback to PATH
    java_exe = shutil.which('java', path=None)
    if java_exe and validate_java(java_exe, min_java_version):
        return pathlib.Path(java_exe)

    # java wasn't found on the system
    return None



import os
import pathlib
import subprocess
import unittest.mock

import pytest

from setuptools_antlr.util import camel_to_snake_case, find_java, validate_java


def test_camel_to_snake_case():
    assert 'ab' == camel_to_snake_case('Ab')
    assert 'ab_cd' == camel_to_snake_case('AbCd')
    assert 'aa_bb_cc' == camel_to_snake_case('AaBbCc')
    assert 'ab0_cd1' == camel_to_snake_case('Ab0Cd1')
    assert 'abcd' == camel_to_snake_case('Abcd')
    assert 'ab' == camel_to_snake_case('AB')
    assert 'ab0' == camel_to_snake_case('AB0')


@unittest.mock.patch('shutil.which')
@unittest.mock.patch('setuptools_antlr.util.validate_java')
def test_find_java_valid_java_home(mock_validate_java, mock_which):
    with unittest.mock.patch.dict('os.environ', {'JAVA_HOME': 'c:/path/to/java'}):
        mock_which.return_value = 'c:/path/to/java/bin/java.exe'
        mock_validate_java.return_value = True

        java_path = find_java('1.7.0')

    assert java_path == pathlib.Path('c:/path/to/java/bin/java.exe')


@unittest.mock.patch('shutil.which')
def test_find_java_invalid_java_home(mock_which):
    with unittest.mock.patch.dict('os.environ', {'JAVA_HOME': 'c:/path/to/java'}):
        mock_which.return_value = None

        java_path = find_java('1.7.0')

    assert java_path is None


@unittest.mock.patch('shutil.which')
def test_find_java_no_java_home(mock_which):
    with unittest.mock.patch.dict('os.environ'):
        del os.environ['JAVA_HOME']
        mock_which.return_value = None

        java_path = find_java('1.7.0')

    assert java_path is None


@unittest.mock.patch('shutil.which')
@unittest.mock.patch('setuptools_antlr.util.validate_java')
def test_find_java_on_path(mock_validate_java, mock_which):
    with unittest.mock.patch.dict('os.environ'):
        del os.environ['JAVA_HOME']
        mock_which.return_value = 'c:/path/to/java/bin/java.exe'
        mock_validate_java.return_value = True

        java_path = find_java('1.7.0')

    assert java_path is not None

test_ids_validate_java = ['valid', 'invalid', 'deprecated', 'corrupt']

test_data_validate_java = [
    (subprocess.CompletedProcess(['java.exe', '-version'], 0, stdout="""
java version "1.8.0_92"
Java(TM) SE Runtime Environment (build 1.8.0_92-b14)
Java HotSpot(TM) 64-Bit Server VM (build 25.92-b14, mixed mode)
"""), True),
        (subprocess.CompletedProcess(['java.exe', '-version'], 0, stdout="""
java version "1.4.0"
Java(TM) 2 Runtime Environment, Standard Edition (build 1.4.0-b03)
Java HotSpot(TM) 64-Bit Server VM (build 1.4.0-b03, mixed mode)
"""), False),
        (subprocess.CompletedProcess(['java.exe', '-version'], 0, stdout="""
java version "1.5.0_22"
Java(TM) 2 Runtime Environment, Standard Edition (build 1.5.0_22-b03)
Java HotSpot(TM) 64-Bit Server VM (build 1.5.0_22-b03, mixed mode)
"""), False),
        (subprocess.CompletedProcess(['java.exe', '-version'], 1, stdout=''), False)
]


@pytest.mark.parametrize('result, expected', test_data_validate_java,
                         ids=test_ids_validate_java)
@unittest.mock.patch('subprocess.run')
def test_validate_java(mock_run, result, expected):
    mock_run.return_value = result

    assert validate_java('java.exe', '1.7.0') == expected



import distutils.errors
import os
import pathlib
import subprocess
import unittest.mock

import pytest
import setuptools.dist

import setuptools_antlr.command
from setuptools_antlr.command import AntlrGrammar, AntlrCommand


@pytest.fixture(scope='module', autouse=True)
def ch_resources_dir(request):
    os.chdir('.')
    init_dir = pathlib.Path.cwd()
    local_dir = os.path.dirname(__file__)
    os.chdir(os.path.join(local_dir, 'resources'))

    def fin():
        os.chdir(str(init_dir))
    request.addfinalizer(fin)


class TestAntlrGrammar:
    def test_read_with_imports(self):
        grammar = AntlrGrammar(pathlib.Path('distributed', 'SomeGrammar.g4'))
        imports = set(grammar.read_imports())

        assert len(imports) == 2
        assert {'CommonTerminals', 'SharedRules'} == imports

    def test_read_without_imports(self):
        grammar = AntlrGrammar(pathlib.Path('distributed', 'CommonTerminals.g4'))
        imports = grammar.read_imports()

        assert not imports

    def test_read_nonexistent_file(self):
        grammar = AntlrGrammar(pathlib.Path('FooBar.g4'))

        # check if DistutilsFileError was thrown
        with pytest.raises(distutils.errors.DistutilsFileError) as excinfo:
            grammar.read_imports()
        assert excinfo.match('FooBar.g4')


class TestAntlrCommand:
    @pytest.fixture(autouse=True)
    def command(self):
        dist = setuptools.dist.Distribution()
        return AntlrCommand(dist)

    @pytest.fixture()
    def configured_command(self, monkeypatch, tmpdir, command):
        command._find_antlr = unittest.mock.Mock(return_value=pathlib.Path('antlr-4.5.3-complete.jar'))
        command._find_grammars = unittest.mock.Mock(return_value=[
            AntlrGrammar(pathlib.Path('standalone/SomeGrammar.g4'))
        ])
        command.output = str(tmpdir.mkdir('gen'))

        monkeypatch.setattr(setuptools_antlr.command, 'find_java',
                            unittest.mock.Mock(return_value=pathlib.Path('c:/path/to/java/bin/java.exe')))

        return command

    test_ids_find_antlr = ['single', 'multiple', 'none', 'invalid']

    test_data_find_antlr = [
        ({'antlr-4.5.3-complete.jar'}, 'antlr-4.5.3-complete.jar'),
        ({'antlr-0.1.1-complete.jar', 'antlr-3.0-complete.jar', 'antlr-4.5.2-complete.jar',
          'antlr-4.5.3-complete.jar'},
         'antlr-4.5.3-complete.jar'),
        ({}, None),
        ({'antlr-runtime-4.5.3.jar'}, None)
    ]

    @pytest.mark.parametrize('available_antlr_jars, expected_antlr_jar', test_data_find_antlr,
                             ids=test_ids_find_antlr)
    def test_find_antlr(self, tmpdir, command, available_antlr_jars, expected_antlr_jar):
        ext_lib_dir = tmpdir.mkdir('lib')

        for jar in available_antlr_jars:
            antlr_jar = ext_lib_dir.join(jar)
            antlr_jar.write('dummy')

        with unittest.mock.patch.object(AntlrCommand, '_EXT_LIB_DIR', str(ext_lib_dir)):
            found_antlr_jar = command._find_antlr()

        assert found_antlr_jar == (pathlib.Path(str(ext_lib_dir), expected_antlr_jar) if expected_antlr_jar
                                   else None)

    test_ids_find_antlr_log = ['single', 'multiple', 'none', 'invalid']

    test_data_find_antlr_log = [
        ({'antlr-2016-12-19-16.01.43.log'}, 'antlr-2016-12-19-16.01.43.log'),
        ({'antlr-2016-12-18-16.01.43.log', 'antlr-2016-12-19-16.01.43.log'}, 'antlr-2016-12-19-16.01.43.log'),
        ({}, None),
        ({'foobar-2016-12-19-16.01.43.log'}, None)
    ]

    @pytest.mark.parametrize('available_antlr_logs, expected_antlr_log', test_data_find_antlr_log,
                             ids=test_ids_find_antlr_log)
    def test_find_antlr_log(self, tmpdir, command, available_antlr_logs, expected_antlr_log):
        package_dir = tmpdir.mkdir('package')

        for log in available_antlr_logs:
            antlr_log = package_dir.join(log)
            antlr_log.write('dummy')

        found_antlr_log = command._find_antlr_log(pathlib.Path(str(package_dir)))

        assert found_antlr_log == (pathlib.Path(str(package_dir), expected_antlr_log) if expected_antlr_log
                                   else None)

    def test_find_grammars_empty(self, tmpdir, command):
        dsl_dir = tmpdir.mkdir('dsl')
        g = command._find_grammars(pathlib.Path(str(dsl_dir)))

        assert len(g) == 0

    def test_find_grammars_standalone(self, command):
        g = command._find_grammars(pathlib.Path('standalone'))

        assert len(g) == 1
        assert g[0].name == 'SomeGrammar'

    def test_find_grammars_distributed(self, command):
        g = command._find_grammars(pathlib.Path('distributed'))

        assert len(g) == 1
        assert g[0].name == 'SomeGrammar'
        d = g[0].dependencies
        assert len(d) == 2
        assert d[0].name == 'CommonTerminals'
        assert d[1].name == 'SharedRules'
        dd = g[0].dependencies[0].dependencies
        assert len(dd) == 0
        dd = g[0].dependencies[1].dependencies
        assert len(dd) == 1
        assert dd[0].name == 'CommonTerminals'

    def test_find_grammars_incomplete(self, command):
        # check if DistutilsFileError was thrown
        with pytest.raises(distutils.errors.DistutilsFileError) as excinfo:
            command._find_grammars(pathlib.Path('incomplete'))
        assert excinfo.match('CommonTerminals')

    def test_finalize_options_default(self, command):
        command.finalize_options()

        assert command.grammars is None
        assert pathlib.Path(command.output) == pathlib.Path('build/lib')
        assert command.atn == 0
        assert command.encoding is None
        assert command.message_format is None
        assert command.long_messages == 0
        assert command.listener == 1
        assert command.visitor == 0
        assert command.depend == 0
        assert command.grammar_options['language'] == 'Python3'
        assert command.w_error == 0
        assert command.x_dbg_st == 0
        assert command.x_dbg_st_wait == 0
        assert command.x_force_atn == 0
        assert command.x_log == 0

    def test_finalize_options_grammars_single(self, command):
        command.grammars = 'Foo'
        command.finalize_options()

        assert 'Foo' in command.grammars

    def test_finalize_options_grammars_multiple(self, command):
        command.grammars = 'Foo Bar'
        command.finalize_options()

        assert 'Foo' in command.grammars
        assert 'Bar' in command.grammars

    def test_finalize_options_grammar_options_language(self, command):
        command.grammar_options = 'language=Python3'
        command.finalize_options()

        assert command.grammar_options['language'] == 'Python3'

    def test_finalize_options_grammar_options_invalid(self, command):
        command.grammar_options = 'language=Java'

        with pytest.raises(distutils.errors.DistutilsOptionError) as excinfo:
            command.finalize_options()
        assert excinfo.match('Java')

    def test_finalize_options_grammar_options_multiple(self, command):
        command.grammar_options = 'superClass=Abc tokenVocab=Lexer'
        command.finalize_options()

        assert command.grammar_options['language'] == 'Python3'
        assert command.grammar_options['superClass'] == 'Abc'
        assert command.grammar_options['tokenVocab'] == 'Lexer'

    def test_finalize_options_debugging_options_invalid(self, capsys, command):
        command.x_dbg_st = 0
        command.x_dbg_st_wait = 1
        command.finalize_options()

        # check if error was logged
        _, err = capsys.readouterr()
        assert 'Waiting for StringTemplate visualizer' in err

    @unittest.mock.patch('setuptools_antlr.command.find_java')
    @unittest.mock.patch('subprocess.run')
    @unittest.mock.patch.object(AntlrCommand, '_find_grammars')
    def test_run_java_found(self, mock_find_grammars, mock_run, mock_find_java, tmpdir, command):
        java_exe = pathlib.Path('c:/path/to/java/bin/java.exe')

        mock_find_java.return_value = java_exe
        mock_find_grammars.return_value = [AntlrGrammar(pathlib.Path('standalone/SomeGrammar.g4'))]
        mock_run.return_value = subprocess.CompletedProcess([], 0)

        command.output = str(tmpdir.mkdir('gen'))
        command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert str(java_exe) in args[0]
        assert '-jar' in args[0]

    @unittest.mock.patch('setuptools_antlr.command.find_java')
    def test_run_java_not_found(self, mock_find_java, command):
        mock_find_java.return_value = None

        with pytest.raises(distutils.errors.DistutilsExecError) as excinfo:
            command.run()
        assert excinfo.match('no compatible JRE')

    @unittest.mock.patch('setuptools_antlr.command.find_java')
    @unittest.mock.patch.object(AntlrCommand, '_find_antlr')
    @unittest.mock.patch('subprocess.run')
    @unittest.mock.patch.object(AntlrCommand, '_find_grammars')
    def test_run_antlr_found(self, mock_find_grammars, mock_run, mock_find_antlr, mock_find_java, tmpdir, command):
        java_exe = pathlib.Path('c:/path/to/java/bin/java.exe')
        antlr_jar = pathlib.Path('antlr-4.5.3-complete.jar')

        mock_find_java.return_value = java_exe
        mock_find_antlr.return_value = antlr_jar
        mock_find_grammars.return_value = [AntlrGrammar(pathlib.Path('standalone/SomeGrammar.g4'))]
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        command.output = str(tmpdir.mkdir('gen'))
        command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert str(antlr_jar) in args[0]

    @unittest.mock.patch('setuptools_antlr.command.find_java')
    @unittest.mock.patch.object(AntlrCommand, '_find_antlr')
    def test_run_antlr_not_found(self, mock_find_antlr, mock_find_java, command):
        java_exe = pathlib.Path('c:/path/to/java/bin/java.exe')

        mock_find_java.return_value = java_exe
        mock_find_antlr.return_value = None

        with pytest.raises(distutils.errors.DistutilsExecError) as excinfo:
            command.run()
        assert excinfo.match('no ANTLR jar')

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_grammars_multiple(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command._find_grammars = unittest.mock.Mock(return_value=[
            AntlrGrammar(pathlib.Path('Foo.g4')),
            AntlrGrammar(pathlib.Path('Bar.g4'))
        ])

        configured_command.grammars = ['Foo']
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.call_count == 1
        assert 'Foo.g4' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_grammars_multiple(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command._find_grammars = unittest.mock.Mock(return_value=[
            AntlrGrammar(pathlib.Path('Foo.g4')),
            AntlrGrammar(pathlib.Path('Bar.g4'))
        ])

        configured_command.grammars = ['Foo', 'Bar']
        configured_command.run()

        assert mock_run.call_count == 2
        args, _ = mock_run.call_args_list[0]
        assert 'Foo.g4' in args[0]
        args, _ = mock_run.call_args_list[1]
        assert 'Bar.g4' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_grammars_not_found(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command._find_grammars = unittest.mock.Mock(return_value=[
            AntlrGrammar(pathlib.Path('Foo.g4'))
        ])

        configured_command.grammars = ['Bar']
        configured_command.run()

        assert mock_run.call_count == 0

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_atn_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.atn = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-atn' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_atn_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.atn = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-atn' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_encoding_specified(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.encoding = 'euc-jp'
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-encoding' in args[0]
        assert 'euc-jp' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_encoding_not_specified(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.encoding = None
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-encoding' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_message_format_specified(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.message_format = 'gnu'
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-message-format' in args[0]
        assert 'gnu' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_message_format_not_specified(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.message_format = None
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-message-format' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_long_messages_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.long_messages = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-long-messages' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_long_messages_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.long_messages = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-long-messages' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_listener_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.listener = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-listener' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_listener_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.listener = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-no-listener' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_visitor_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.visitor = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-visitor' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_visitor_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.visitor = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-no-visitor' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_grammar_options_specified(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.grammar_options = {'superClass': 'Foo', 'tokenVocab': 'Bar'}
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-DsuperClass=Foo' in args[0]
        assert '-DtokenVocab=Bar' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_grammar_options_not_specified(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.grammar_options = {}
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert not any(a.startswith('-D') for a in args[0])

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_depend_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0, stdout='FooParser.py : Foo.g4')

        configured_command.depend = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-depend' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_depend_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)
        configured_command.depend = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-depend' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_w_error_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.w_error = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-Werror' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_w_error_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.w_error = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-Werror' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_dbg_st_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_dbg_st = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-XdbgST' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_dbg_st_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_dbg_st = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-XdbgST' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_dbg_st_wait_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_dbg_st_wait = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-XdbgSTWait' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_dbg_st_wait_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_dbg_st_wait = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-XdbgSTWait' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_force_atn_wait_enabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_force_atn = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-Xforce-atn' in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_force_atn_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_force_atn = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-Xforce-atn' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    @unittest.mock.patch.object(AntlrCommand, '_find_antlr_log')
    @unittest.mock.patch('shutil.move')
    def test_run_x_log_enabled(self, mock_move, mock_find_antlr_log, mock_run, capsys, configured_command):
        log_file = 'antlr-2016-12-19-16.01.43.log'
        mock_run.return_value = unittest.mock.Mock(returncode=0)
        mock_find_antlr_log.return_value = pathlib.Path(log_file)

        configured_command.x_log = 1
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-Xlog' in args[0]

        args, _ = mock_move.call_args
        assert mock_move.called
        assert log_file in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_x_log_disabled(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.x_log = 0
        configured_command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-Xlog' not in args[0]

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    @unittest.mock.patch.object(AntlrCommand, '_find_antlr_log')
    def test_run_x_log_not_found(self, mock_find_antlr_log, mock_run, capsys, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)
        mock_find_antlr_log.return_value = None

        configured_command.x_log = 1
        configured_command.run()

        _, err = capsys.readouterr()
        assert 'no logging info dumped out by ANTLR' in err

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_parser_generation_successful(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        configured_command.run()

        args, kwargs = mock_run.call_args
        assert mock_run.called
        assert 'SomeGrammar.g4' in args[0]
        assert kwargs['cwd'] == 'standalone'

    @pytest.mark.usefixtures('configured_command')
    @unittest.mock.patch('subprocess.run')
    def test_run_parser_generation_failed(self, mock_run, configured_command):
        mock_run.return_value = unittest.mock.Mock(returncode=-1)

        with pytest.raises(distutils.errors.DistutilsExecError) as excinfo:
            configured_command.run()
        assert excinfo.match('SomeGrammar parser couldn\'t be generated')

    @unittest.mock.patch('setuptools_antlr.command.find_java')
    @unittest.mock.patch.object(AntlrCommand, '_find_antlr')
    @unittest.mock.patch('subprocess.run')
    @unittest.mock.patch.object(AntlrCommand, '_find_grammars')
    def test_run_one_library_location(self, mock_find_grammars, mock_run, mock_find_antlr, mock_find_java, tmpdir,
                                      command):
        java_exe = pathlib.Path('c:/path/to/java/bin/java.exe')
        antlr_jar = pathlib.Path('antlr-4.5.3-complete.jar')
        mock_find_java.return_value = java_exe
        mock_find_antlr.return_value = antlr_jar

        grammar = AntlrGrammar(pathlib.Path('SomeGrammar.g4'))
        grammar.dependencies = [
            AntlrGrammar(pathlib.Path('shared/CommonTerminals.g4'))
        ]
        mock_find_grammars.return_value = [grammar]
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        command.output = str(tmpdir.mkdir('gen'))
        command.run()

        args, _ = mock_run.call_args
        assert mock_run.called
        assert '-lib' in args[0]
        assert any(a.endswith('shared') for a in args[0])

    @unittest.mock.patch('setuptools_antlr.command.find_java')
    @unittest.mock.patch.object(AntlrCommand, '_find_antlr')
    @unittest.mock.patch('subprocess.run')
    @unittest.mock.patch.object(AntlrCommand, '_find_grammars')
    def test_run_multiple_library_location(self, mock_find_grammars, mock_run, mock_find_antlr, mock_find_java, tmpdir,
                                           command):
        java_exe = pathlib.Path('c:/path/to/java/bin/java.exe')
        antlr_jar = pathlib.Path('antlr-4.5.3-complete.jar')
        mock_find_java.return_value = java_exe
        mock_find_antlr.return_value = antlr_jar

        grammar = AntlrGrammar(pathlib.Path('SomeGrammar.g4'))
        grammar.dependencies = [
            AntlrGrammar(pathlib.Path('terminals/common.g4')),
            AntlrGrammar(pathlib.Path('rules/common.g4'))
        ]
        mock_find_grammars.return_value = [grammar]
        mock_run.return_value = unittest.mock.Mock(returncode=0)

        command.output = str(tmpdir.mkdir('gen'))

        with pytest.raises(distutils.errors.DistutilsOptionError) as excinfo:
            command.run()
        assert excinfo.match('Imported grammars of \'SomeGrammar\' are located in more than one directory.')



"""Setup script for setuptools-antlr."""
import sys

import setuptools

# need to guard script here due to reentrance while testing multiprocessing:
if __name__ == '__main__':
    if sys.version_info < (3, 5):
        print('This Python version isn\'t supported, minimal version 3.5 is required.')
        exit(1)

    needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
    pytest_runner_opt = ['pytest-runner>=2.0,<3dev'] if needs_pytest else []

    needs_pylint = {'lint'}.intersection(sys.argv)
    setuptools_lint_opt = ['setuptools-lint>=0.4,<0.5dev'] if needs_pylint else []

    needs_flake8 = {'flake8'}.intersection(sys.argv)
    flake8_opt = ['flake8>=2.0,<3dev', 'flake8-import-order>=0.9,<0.10dev',
                  'flake8-todo>=0.5,<0.6dev', 'pep8-naming>=0.3,<0.4dev'] if needs_flake8 else []

    setuptools.setup(
        name='setuptools-antlr',
        version='0.0.5',
        packages=setuptools.find_packages(),
        package_data={'setuptools_antlr': ['lib/antlr-4.6-complete.jar', 'lib/LICENSE.txt']},
        entry_points={
            'distutils.commands': [
                'antlr = setuptools_antlr.command:AntlrCommand'
            ]
        },
        tests_require=[
            'pytest-cov'
        ],
        setup_requires=pytest_runner_opt + setuptools_lint_opt + flake8_opt,
        url='https://github.com/ferraith/setuptools-antlr',
        license='MIT',
        author='Andreas Schmidl',
        author_email='Andreas.Schmidl@gmail.com',
        description='Setuptools command for generating ANTLR based parsers.',
        platforms=[
            'any'
        ],
        keywords='antlr setuptools dsl',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'Topic :: Software Development',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6'
        ]
    )
